import re

# Top level patterns for all commands. Used to recursively parse script.
PATTERN_STRINGS = {
    # "env_variable": r"""(?:[^\s\*][^\n]*)?(\%(\S+?)\%)""",
    "env_variable": r"""\%(\S+?)\%""",

    "user_function": r"""
        @(\S+?)  # Function name
        \(
            ([^)]*?)  # Arguments
        \)
    """,

    "set": r"""

                \$set(global|local|)
                \s*
                (\S+)
                \s+
                ([^\s\;]+)
            """,

    "eval": r"""

                \$eval(global|local|)
                \s*
                (\S+)
                \s+
                ([^\;]+)
            """,

    "import": r"""
        \$Import
        \s+
        ([^\s;]+)                # File name
    """,

    "block": r"\$Block\s+(.+?)\s+(.*?)\$EndBlock",

    "group": r"\$Group\s+(.+?)\s+(.*?;)",  # 1) name, 2) content

    "pgroup": r"\$PGroup\s+(.+?)\s+(.*?;)",

    "display": r"""
        \$Display\s
            (.+?;)      # Variables and groups to be displayed
    """,

    "display_all": r"""
        \$Display_all
            (.+?;)      # Variables and groups to be displayed
    """,

    "model": r"\$Model\s+(.+?)\s+(.*?);",

    "solve": r"\$Solve\s+(.*?);",

    "fix": r"""
                        (\$(?:UN)?FIX)          # Fix or unfix command
                        (?:[(\[] (
                            -? (?:\d+\.?\d*|INF|EPS) (?: \,\s*-?(?:\d+\.?\d*|INF|EPS) )?       # Optional arguments
                        ) [)\]])?
                        \s+
                        (.+?;)                  # Content
                    """,


    "if": r"""
                    \$If(?P<if_id>\d*)\s+
                    ([^:]*?)             # Condition
                    \:
                    (.*?)
                    \$EndIF(?P=if_id)\b
                """,

    "define_function": r"""

                        \$Function(?P<function_id>\d*)\s+
                        (\S+?)        # Name $1
                        [(\[]([^)\]]*)[)\]]  # Arguments  $2
                        (?:\:)?\s+
                        (.*?)         # Expression $3
                        \$EndFunction(?P=function_id)\b
    """,

    "for_loop": r"""

                        \$For(?P<for_id>\d*)
                        \s+
                        (.+?) # The iterator
                        \s+
                        in
                        \s+
                        ([^\:]*?)  # The iterable
                        :
                        (.*?)
                        \$EndFor(?P=for_id)\b
    """,


    "loop": r"""
                        \$Loop(?P<loop_id>\d*)
                        \s*
                        (.*?)               # Group name
                        [:]
                        (.*?)
                        \$EndLoop(?P=loop_id)\b
                    """,
    "replace": r"""
                        \$Replace
                        [(\[]
                        ('.*?'|".*?")       # String to find
                        ,\s*
                        ('.*?'|".*?")       # Replacement string
                        (,\s*\d+)?          # Max replacements
                        [)\]]
                        (.*?)
                        \$EndReplace
                    """,
    "regex": r"""
                        \$Regex(?P<regex_id>\d*)
                        [(\[]\s*
                        ('.*?'|".*?"|[^,]+)       # String to find
                        ,\s*
                        ('.*?'|".*?")             # Replacement string
                        (,\s*\d+)?                # Max replacements
                        [)\]]
                        (.*?)
                        \$EndRegex(?P=regex_id)\b
                    """,

}

# Compile regex patterns
PATTERNS = {k: re.compile(v, re.VERBOSE | re.IGNORECASE | re.MULTILINE | re.DOTALL) for k, v in PATTERN_STRINGS.items()}
# Create combined pattern that matches any of the patterns
PATTERNS["Any"] = re.compile("|".join(PATTERN_STRINGS.values()), re.VERBOSE | re.IGNORECASE | re.MULTILINE | re.DOTALL)


PATTERNS["TopDown"] = re.compile("|".join((  # Remember to also add these to the list in gamY
        PATTERN_STRINGS["if"],
        PATTERN_STRINGS["for_loop"],
        PATTERN_STRINGS["loop"],
        PATTERN_STRINGS["define_function"],
    )),
    re.VERBOSE | re.IGNORECASE | re.MULTILINE | re.DOTALL
)



import os
import sys
sys.path.insert(0, os.getcwd())

import dreamtools as dt
import pandas as pd
import plotly.express as px
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import numpy as np

# Dette er muligvis nødvendigt afhængingt af installationen
pd.options.plotting.backend = "plotly"
import plotly.io as pio
pio.renderers.default = "browser"

# Indlæs banker
dt.REFERENCE_DATABASE = dt.Gdx("C:/Users/B031441/Desktop/Match/Model/Gdx/dynamic_calibration2.gdx")
b = dt.Gdx("C:/Users/B031441/Desktop/Match/Model/Gdx/dynamic_calibration2.gdx")

b = dt.Gdx("C:/Users/B031441/Desktop/Match/Model/Gdx/smooth_profiles_input.gdx")
c = dt.Gdx("C:/Users/B031441/Desktop/Match/Model/Gdx/smooth_profiles.gdx")

pd.DataFrame({"a.uArv": a.uArv, "b.uArv": b.uArv, "c.uArv": c.uArv}).plot()

def age3d(series, t_start=2016, t_end=2099, a_start=0, a_end=100, title="", ztitle=""):
  age = list(range(a_start, a_end+1))
  time = list(range(t_start, t_end+1))
  value = series.loc[age, time].unstack().values
  surface = go.Surface(x=time, y=age, z=value)
  return go.Figure(
    surface,
    layout={
      "scene": {
        "xaxis": {"title": "År", "autorange": "reversed"},
        "yaxis": {"title": "Alder", "autorange": "reversed"},
        "zaxis": {"title": ztitle},
      },
      "title": {"text": title, 'x': 0.5, "y": 0.925}
    }
  )

figures = [
  age3d(c.vC_a, 2015, 2099, a_end=99, title="Forbrug (vC_a)", ztitle="Mio. kr. (vækst- og inflations-korrigeret)"),
  age3d(c.vHhx, 2015, 2099, a_end=99, title="Formue ekskl. pension, bolig og realkreditgæld (vHhx)", ztitle="Mio. kr. (vækst- og inflations-korrigeret)"),
  age3d(c.vBolig, 2015, 2099, a_end=99, title="Bolig-formue (vBolig)", ztitle="Mio. kr. (vækst- og inflations-korrigeret)"),
  age3d(c.vHh["Net"], 2015, 2099, a_end=99, title="Netto finansiel formue (vHh[Net])",
        ztitle="Mio. kr. (vækst- og inflations-korrigeret)"),
]
figures_to_html(figures, "aldersprofiler2.html")

v = "rDisk"
df = pd.DataFrame({
  # f"a.{v}": a[v].sort_index().loc[0:100, [2017, 2040]],
  f"b.{v}": b[v].sort_index().loc[0:100, [2017, 2065]],
  f"c.{v}": c[v].sort_index().loc[0:100, [2017, 2065]],
}).unstack()
df.columns = ["".join(str(i)) for i in df]
df.plot(title=v)

def debeta(beta):
  return 1/beta - 1 - 0.1

from scipy.ndimage import gaussian_filter1d
from scipy.signal import savgol_filter, gauss_spline
from scipy.interpolate import UnivariateSpline

def spline(x, y, s=None, k=3, padding=1):
  if s is None:
    s = len(x) * np.std(y[1:]/y[:-1].values) ** 2
  padded = [*[x[0] for _ in range(padding)], *x, *[x[-1] for _ in range(padding)]]
  return pd.Series(UnivariateSpline(padded, y[padded], k=k, s=s)(x), x)

figs = []
b.rDisk_a = 1/(1+b.rDisk_a)
for var_name, a_start, smoothness in [
  ("rDisk_a", 18, 3),
  # ("uBolig_a", 18, 4),
  # #  ("uBoligHtM_a", 18, 0.3),
  # ("fProdHh_a", 15, 3),
  # ("ftBund", 15, 3),
  # ("ftKommune", 15, 3),
  # ("rTopSkatInd", 15, 3),
  # ("fvPersInd_a", 15, 2),
  # # ("cHh_a", 0, 4),
  # ("rRealKred2Bolig_a", 18, 2),
  # ("rvCLejeBolig", 18, 3),
  # ("uBoernFraHh_a", 0, 3),
]:
  x = np.array(range(a_start,101))
  df = {}
  for t in [2017]:
    y = 1/(1+b[var_name].sort_index().loc[a_start:100].loc[:,t])
    df[f"{var_name}[{t}]"] = y
    df[f"gaussian[{t}], sigma={smoothness}"] = pd.Series(gaussian_filter1d(y, smoothness, mode="nearest"), x)
    # df[f"savgol[{t}]"] = pd.Series(savgol_filter(y, 11, 2, mode="nearest"), x)
    # df[f"MA[{t}]"] = y.rolling(5, min_periods=0).mean()
    # df[f"UnivariateSpline[{t}], k=3, s={smoothness}"] = pd.Series(UnivariateSpline(x, y, k=3, s=smoothness)(x), x)
    # df[f"UnivariateSpline[{t}], k=5, s={smoothness}"] = pd.Series(UnivariateSpline(x, y, k=5, s=smoothness)(x), x)
    # df[f"Spline, s={smoothness}"] = spline(x, y, smoothness)
    # df["Spline, padding=0, k=3"] = spline(x, y, padding=0, k=3)
    # df["Spline, padding=0, k=5"] = spline(x, y, padding=0, k=5)
    # df["Spline, padding=1, k=3"] = spline(x, y, padding=1, k=3)
    # df["Spline, padding=10, k=3"] = spline(x, y, padding=10, k=3)
    # df["Spline, s=std*span"] = spline(x, y, padding=0)
    # df["gauss_spline"] = pd.Series(gauss_spline(y.values, 3), x)
  figs.append(px.line(df, title=var_name))
figures_to_html(figs)




def figures_to_html(figs, filename="dashboard.html"):
  dashboard = open(filename, 'w')
  dashboard.write("<html><head></head><body>" + "\n")
  for fig in figs:
    inner_html = fig.to_html().split('<body>')[1].split('</body>')[0]
    dashboard.write(inner_html)
  dashboard.write("</body></html>" + "\n")




figures = []
for i in ["Alder", "Kap", "Pens", "PensX"]:
  figures.append(age3d(a.nPop * a.vPensIndb.loc[i], title=f"Indbetaling, {i}"))
  figures.append(age3d(a.nPop * a.vPensUdb.loc[i], title=f"Udbetaling, {i}"))
figures_to_html(figures, "pensions.html")


r = b.rDisk_a[b.rDisk_a != 0][:,2017]
px.line({
  "gaussian": gaussian_filter1d(1/(1+r), 5, mode="nearest"),
  # "beta_first_s=3": gaussian_filter1d(1/(1+r), 3, mode="nearest"),
  "golsov": savgol_filter(1/(1+r), 21, 2, mode="nearest"),
  "smooth_first": (1/(1+gaussian_filter1d(r, 5, mode="nearest"))),
  "beta": 1/(1+r),
}).show()


from statsmodels.nonparametric.smoothers_lowess import lowess
from scipy.interpolate import UnivariateSpline
from scipy.signal import savgol_filter, wiener, medfilt
from scipy.ndimage import gaussian_filter1d, gaussian_laplace

figs = []
y = 1/(1+b.rDisk).loc[18:100].loc[:,2017]
x = b.rDisk.index.values
var_name="rDisk"
df = {
  var_name: 1/y - 1.1,
  # "s=0.005": UnivariateSpline(x, y, k=3, s=0.005)(x),
  # "s=0.01": UnivariateSpline(x, y, k=3, s=0.01)(x),
  # "s=0.1": UnivariateSpline(x, y, k=3, s=0.1)(x),
  # "s=1": UnivariateSpline(x, y, k=3, s=1)(x),
  # "s=100": UnivariateSpline(x, y, k=3, s=100)(x),
  # "s=std*span": UnivariateSpline(x, y, k=3, s=np.std(y)*(max(y)-min(y)))(x),
  # "savgol, 15": savgol_filter(y, 15, 3, mode="nearest"),
  # "gaussian 1": gaussian_filter1d(y, 1, mode="nearest"),
  # "gaussian 2": gaussian_filter1d(y, 2, mode="nearest"),
  "gaussian 3, nearest": 1/gaussian_filter1d(y, 3, mode="nearest")-1.1,
  "gaussian 3": 1/gaussian_filter1d(y, 3)-1.1,
  "gaussian 3, 19+": 1/(np.append(y.values[0], gaussian_filter1d(y.loc[19:], 3)))-1.1,
  # "gaussian 4": gaussian_filter1d(y, 4, mode="nearest"),
  # f"gaussian, sigma={smoothness}": gaussian_filter1d(y, smoothness, mode="nearest"),
  # "polynomial(deg=7)": np.polynomial.Polynomial.fit(x, y, 7).linspace(83)[1],
  # "lowess(frac=0.1)": lowess(y, x, frac=0.1, return_sorted=False),
}
figs.append(px.line(df, title=var_name))
figures_to_html(figs)




PLOTLY = True
try:
  import plotly
except ImportError:
  PLOTLY = False

if PLOTLY:
  import pandas
  pandas.options.plotting.backend = "plotly"
  from .timeseries_analysis import *
  from .plotly_util import *
  from .age_plot import *



from .gams_pandas import *
from .multiindex_plotly import *

# Global setting controlling the default position of the time index (-1 = last index is time)
X_AXIS_INDEX = -1

# Time settings, used for plot and prt functions
START_YEAR = 1965
END_YEAR = 2100

# Time settings, currently used for plotting only
START_AGE = 0
END_AGE = 100

# Global databases
REFERENCE_DATABASE = None

# Domain specific settings
from .multiindex_plotly import makro_settings as settings
DEFAULT_SET_AGGREGATIONS = settings.DEFAULT_SET_AGGREGATIONS
AGE_AXIS_TITLE = settings.AGE_AXIS_TITLE
TIME_AXIS_TITLE = settings.TIME_AXIS_TITLE



"""
"""
import os
from collections.abc import Iterable
from six import string_types

import numpy as np
import pandas as pd
import gams
import logging
import builtins
from copy import deepcopy

logger = logging.getLogger(__name__)

def is_iterable(arg):
  return isinstance(arg, Iterable) and not isinstance(arg, string_types)

def map_lowest_level(func, x):
  """Map lowest level of zero or more nested lists."""
  if is_iterable(x):
    return [map_lowest_level(func, i) for i in x]
  else:
    return func(x)

def try_to_int(x):
  try:
    if str(int(x)) == str(x):
      return int(x)
    else:
      return x
  except ValueError:
    return x

def map_to_int_where_possible(iter):
  """Returns an iterable where each element is converted to an integer if possible for that element."""
  return map_lowest_level(try_to_int, iter)


class GamsPandasDatabase:
  """
  GamsPandasDatabase converts sets, parameters, and variables between a GAMS database and Pandas series.
  When as symbol is first retrieved it is converted to a Pandas series and stored in self.series
  Changes to retrieved series are written to the GAMS database on export.
  """
  def __init__(self, database=None, workspace=None):
    if database is None:
      if workspace is None:
        workspace = gams.GamsWorkspace()
      database = workspace.add_database()
    self.database = database
    self.series = {}

  def __getattr__(self, item):
    return self[item]

  def copy(self):
    obj = type(self).__new__(self.__class__)
    obj.__dict__.update(self.__dict__)
    obj.database = self.database.workspace.add_database(source_database=self.database)
    obj.series = deepcopy(self.series)
    return obj

  def merge(self, other, symbol_names=None, inplace=False):
    """
    Merge two GamsPandasDatabases.
    symbol_names: list of symbol names to get from other. If None, all symbols in other are merged.
    """
    if inplace:
      db = self
    else:
      db = self.copy()

    if symbol_names is None:
      symbol_names = other.symbols

    for name in other.sets:
      if name in symbol_names:
        db.add_set_from_series(other[name].texts, other[name].explanatory_text)

    for name in other.variables:
      if name in symbol_names:
        db.add_variable_from_series(other[name], other[name].explanatory_text)

    for name in other.parameters:
      if name in symbol_names:
        db.add_parameter_from_series(other[name], other[name].explanatory_text)

    return db

  @property
  def symbols(self):
    """Dictionary of all symbols in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database}

  @property
  def sets(self):
    """Dictionary of all sets in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsSet)}

  @property
  def variables(self):
    """Dictionary of all variables in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsVariable)}

  @property
  def parameters(self):
    """Dictionary of all parameters in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsParameter)}

  @property
  def equations(self):
    """Dictionary of all equations in the underlying GAMS database"""
    return {symbol.name: symbol for symbol in self.database if isinstance(symbol, gams.GamsEquation)}

  def add_to_builtins(self, *args):
    """Retrieve any number symbol names from the database and add their Pandas representations to the global namespace."""
    for identifier in args:
      setattr(builtins, identifier, self[identifier])

  def get(self, *args):
    """Retrieve any nymber of symbol names and return a list of their Pandas representations."""
    return [self[i] for i in args]

  def add_set_from_series(self, series, explanatory_text=""):
    """Add set symbol to database based on a Pandas series."""
    self.create_set(series.name, series.index, explanatory_text, texts=series)

  def add_parameter_from_series(self, series, explanatory_text="", add_missing_domains=False):
    """Add parameter symbol to database based on a Pandas series."""
    if len(series) == 1:
      df = pd.DataFrame(series)
    else:
      df = series.reset_index()
    self.add_parameter_from_dataframe(series.name, df, explanatory_text, add_missing_domains)

  def add_parameter_from_dataframe(self, identifier, df, explanatory_text="", add_missing_domains=False, value_column_index=-1):
    """Add parameter symbol to database based on a Pandas DataFrame."""
    domains = list(df.columns[:value_column_index:])
    for d in domains:
      if d not in self:
        if add_missing_domains:
          self.add_set_from_series(df[d])
        else:
          raise KeyError(f"'{d}' is not a set in the database. Enable add_missing_domains or add the set to the database manually.")
    self.database.add_parameter_dc(identifier, domains, explanatory_text)
    self.series[identifier] = df.set_index(domains).iloc[:, 0]

  def add_variable_from_series(self, series, explanatory_text="", add_missing_domains=False):
    """Add variable symbol to database based on a Pandas series."""
    if len(series) == 1:
      df = pd.DataFrame(series)
    else:
      df = series.reset_index()
    self.add_variable_from_dataframe(series.name, df, explanatory_text, add_missing_domains)

  def get_index(self, x):
    if x is None or isinstance(x, pd.Index):
      return x
    elif isinstance(x, str):
      return self[x]
    elif len(x) and isinstance(x[0], (pd.Index, tuple, list)):
      multi_index = pd.MultiIndex.from_product(x)
      multi_index.names = [getattr(i, "name", None) for i in x]
      return multi_index
    else:
      return pd.Index(x)

  def get_domains_from_index(self, index, name):
    if hasattr(index, "domains"):
      domains = index.domains
    elif hasattr(index, "name"):
      domains = index.names
    else:
      domains = [index.name]
    return ["*" if i in (None, name) else i for i in domains]

  def create_set(self, name, index, explanatory_text="", texts=None, domains=None):
    """
    Add a new GAMS Set to the database and return an Pandas representation of the Set.
    :param str name: Name of the set
    :param iterable index: Iterable of record keys to be added to the set
    :param str explanatory_text: Explanatory text added to the GAMS set
    :param iterable texts: Iterable of record labels - should match the size of the index parameter.
    :param iterable domains: Names of domains that the set should be defined over
    :return: Pandas Index
    """
    if len(index) and isinstance(index[0], pd.Index):
      multi_index = pd.MultiIndex.from_product(index)
      multi_index.names = [getattr(i, "name", None) for i in index]
      index = multi_index
    elif isinstance(index, pd.Index):
      index = index.copy()
    else:
      index = pd.Index(index)

    index.explanatory_text = explanatory_text

    if texts is None:
      texts = map_lowest_level(str, index)
    index.texts = pd.Series(texts, index=index)
    index.texts.name = index.name

    if domains is None:
      domains = ["*" if i in (None, name) else i for i in self.get_domains_from_index(index, name)]
    index.domains = domains
    index.names = domains
    index.name = name

    self.database.add_set_dc(index.name, domains, explanatory_text)
    self.series[index.name] = index
    return self[name]

  def create_variable(self, name, index=None, explanatory_text="", data=None, dtype=None, copy=False, add_missing_domains=False):
    if index is not None:
      series = pd.Series(data, self.get_index(index), dtype, name, copy)
      series.explanatory_text = explanatory_text
      self.add_variable_from_series(series, explanatory_text, add_missing_domains)
    elif isinstance(data, pd.DataFrame):
      self.add_variable_from_dataframe(name, data, explanatory_text, add_missing_domains)
    elif isinstance(data, pd.Series):
      self.add_variable_from_series(data, explanatory_text, add_missing_domains)
    else:
      if is_iterable(data) and len(data) and is_iterable(data[0]):
        self.database.add_variable(name, len(data[0]), gams.VarType.Free, explanatory_text)
      elif is_iterable(data):
        self.database.add_variable(name, 1, gams.VarType.Free, explanatory_text)
      else:
        self.database.add_variable(name, 0, gams.VarType.Free, explanatory_text)
      self[name] = data
    return self[name]

  def create_parameter(self, name, index=None, explanatory_text="", data=None, dtype=None, copy=False, add_missing_domains=False):
    if index is not None:
      series = pd.Series(data, self.get_index(index), dtype, name, copy)
      series.explanatory_text = explanatory_text
      self.add_parameter_from_series(series, explanatory_text, add_missing_domains)
    elif isinstance(data, pd.DataFrame):
      self.add_parameter_from_dataframe(name, data, explanatory_text, add_missing_domains)
    elif isinstance(data, pd.Series):
      self.add_parameter_from_series(data, explanatory_text, add_missing_domains)
    else:
      if is_iterable(data) and len(data) and is_iterable(data[0]):
        self.database.add_parameter(name, len(data[0]), explanatory_text)
      elif is_iterable(data):
        self.database.add_parameter(name, 1, explanatory_text)
      else:
        self.database.add_parameter(name, 0, explanatory_text)
      self[name] = data
    return self[name]

  def add_variable_from_dataframe(self, identifier, df, explanatory_text="", add_missing_domains=False, value_column_index=-1):
    """Add variable symbol to database based on a Pandas DataFrame."""
    domains = list(df.columns[:value_column_index:])
    for d in domains:
      if d not in self:
        if add_missing_domains:
          self.add_set_from_series(df[d])
        else:
          raise KeyError(f"'{d}' is not a set in the database. Enable add_missing_domains or add the set to the database manually.")
    self.database.add_variable_dc(identifier, gams.VarType.Free, domains, explanatory_text)
    self.series[identifier] = df.set_index(domains).iloc[:, 0]

  @staticmethod
  @np.vectorize
  def detuple(t):
    """Returns the iterable unchanged, except if it is a singleton, then the element is returned"""
    if isinstance(t, str):
      return t
    try:
      if len(t) == 1:
        return t[0]
    except TypeError:
      pass
    return t

  def __getitem__(self, item):
    if item not in self.series:
      symbol = self.symbols[item]

      if isinstance(symbol, gams.GamsSet):
        self.series[item] = index_from_symbol(symbol)

      elif isinstance(symbol, gams.GamsVariable):
        if symbol_is_scalar(symbol):
          self.series[item] = symbol.find_record().level if len(symbol) else None
        else:
          self.series[item] = series_from_variable(symbol)

      elif isinstance(symbol, gams.GamsParameter):
        if symbol_is_scalar(symbol):
          self.series[item] = symbol.find_record().value if len(symbol) else None
        else:
          self.series[item] = series_from_parameter(symbol)

      elif isinstance(symbol, gams.GamsEquation):
        return symbol

    return self.series[item]

  def __setitem__(self, name, value):
    if name in self.symbols:
      if not is_iterable(value) and is_iterable(self[name]):  # If assigning a scalar to all records in a series
        value = pd.Series(value, index=self[name].index)
      set_symbol_records(self.symbols[name], value)
      self.series[name] = value
    else:
      if not value.name:
        value.name = name
      self.series[name] = value

  def items(self):
    return self.symbols.items()

  def keys(self):
    return self.symbols.keys()

  def values(self):
    return self.symbols.values()

  def save_series_to_database(self):
    """Save Pandas series to GAMS database"""
    for symbol_name, series in self.series.items():
      set_symbol_records(self.symbols[symbol_name], series)

  def export(self, path):
    """Save changes to database and export database to GDX file."""
    self.save_series_to_database()
    self.database.export(os.path.abspath(path))

  def __iter__(self):
    return iter(self.symbols)

  def __len__(self):
    return len(self.symbols)

  def get_text(self, name):
    """Get explanatory text of GAMS symbol."""
    return self.symbols[name].get_text()


def merge_symbol_records(series, symbol):
  """Convert Pandas series to records in a GAMS Symbol"""
  if isinstance(symbol, gams.GamsSet):
    attr = "text"
  elif isinstance(symbol, gams.GamsVariable):
    attr = "level"
  elif isinstance(symbol, gams.GamsParameter):
    attr = "value"
  for k, v in series.items():
    setattr(symbol.merge_record(k), attr, v)

def set_symbol_records(symbol, value):
  """Convert Pandas series to records in a GAMS Symbol"""
  if isinstance(symbol, gams.GamsSet):
    if isinstance(value, pd.Index):
      texts = getattr(value, "texts", None)
      value = texts if texts is not None else pd.Series(map(str, value), index=value)
    def add_record(symbol, k, v):
      if not all_na(v):
        symbol.add_record(k).text = str(v)
  elif isinstance(symbol, gams.GamsVariable):
    def add_record(symbol, k, v):
      if not all_na(v):
        symbol.add_record(k).level = v
  elif isinstance(symbol, gams.GamsParameter):
    def add_record(symbol, k, v):
      if not all_na(v):
        symbol.add_record(k).value = v
  else:
    TypeError(f"{type(symbol)} is not (yet) supported by gams_pandas")

  symbol.clear()
  if symbol_is_scalar(symbol):
    add_record(symbol, None, value)
  elif list(value.keys()) == [0]:  # If singleton series
    add_record(symbol, None, value[0])
  else:
    for k, v in value.items():
      add_record(symbol, map_lowest_level(str, k), v)

def all_na(x):
  """Returns bool of whether a series or scalar consists of all NAs"""
  if is_iterable(x):
    return all(pd.isna(x))
  else:
    return pd.isna(x)

def index_names_from_symbol(symbol):
  """
  Return the domain names of a GAMS symbol,
  except ['*'] cases are replaced by the name of the symbol
  and ['*',..,'*'] cases are replaced with ['index_0',..'index_n']
  """
  index_names = list(symbol.domains_as_strings)
  if index_names == ["*"]:
    return [symbol.name]
  if index_names.count("*") > 1:
    for i, name in enumerate(index_names):
      if name == "*":
        index_names[i] = f"index_{i}"
  return index_names

def index_from_symbol(symbol):
  """Return a Pandas Index based on the records and domain names of a GAMS symbol."""
  if len(symbol.domains_as_strings) > 1:
    keys = map_to_int_where_possible([rec.keys for rec in symbol])
    index = pd.MultiIndex.from_tuples(keys, names=index_names_from_symbol(symbol))
    index.name = symbol.name
  elif len(symbol.domains_as_strings) == 1:
    keys = map_to_int_where_possible([rec.keys[0] for rec in symbol])
    index = pd.Index(keys, name=index_names_from_symbol(symbol)[0])
  else:
    return None
  if isinstance(symbol, gams.GamsSet):
    index.text = symbol.text
    index.domains = symbol.domains_as_strings
    index.texts = pd.Series([rec.text for rec in symbol], index, name=symbol.name)
  return index

def symbol_is_scalar(symbol):
  return not symbol.domains_as_strings

def series_from_variable(symbol, attr="level"):
  """Get a variable symbol from the GAMS database and return an equivalent Pandas series."""
  return pd.Series([getattr(rec, attr) for rec in symbol], index_from_symbol(symbol), name=symbol.name)

def series_from_parameter(symbol):
  """Get a parameter symbol from the GAMS database and return an equivalent Pandas series."""
  return pd.Series([rec.value for rec in symbol], index_from_symbol(symbol), name=symbol.name)

class Gdx(GamsPandasDatabase):
  """Wrapper that opens a GDX file as a GamsPandasDatabase."""
  def __init__(self, file_path, workspace=None):
    if not file_path:
      raise ValueError("No file path was provided")
      return None
    if not os.path.splitext(file_path)[1]:
      file_path = file_path + ".gdx"
    self.abs_path = os.path.abspath(file_path)
    logger.info(f"Open GDX file from {self.abs_path}.")
    if workspace is None:
      workspace = gams.GamsWorkspace()
    database = workspace.add_database_from_gdx(self.abs_path)
    super().__init__(database)

  def export(self, path=None, relative_path=True):
    """
    Save changes to database and export database to GDX file.
    If no path is specified, the path of the linked GDX file is used, overwriting the original.
    Use relative_path=True to use export to a path relative to the directory of the originally linked GDX file.
    """
    if path is None:
      path = self.abs_path
    elif relative_path:
      path = os.path.join(os.path.dirname(self.abs_path), path)
    logger.info(f"Export GDX file to {path}.")
    super().export(path)



import dreamtools as dt
import numpy as np
import pandas as pd
import easygui
from IPython.display import display

def time(start, end=None):
  """Set global time settings."""
  if end is None:
    end = start
  dt.START_YEAR = start
  dt.END_YEAR = end

def aggregate_index(index, default_set_aggregations):
  if index.nlevels > 1:
    return tuple(default_set_aggregations.get(k, list(index.levels[i])) for i, k in enumerate(index.names))
  else:
    return default_set_aggregations.get(index.name, list(index))

def aggregate_series(series, default_set_aggregations=None, reference_database=None):
  if default_set_aggregations is None:
    default_set_aggregations = dt.DEFAULT_SET_AGGREGATIONS

  if reference_database is None:
    reference_database = get_reference_database()

  if len(reference_database[series.name].index) == len(series.index):
    return series.loc[aggregate_index(series.index, default_set_aggregations)]
  else:
    return series

def prt(iter_series,
        operator=None,
        start_year=None,
        end_year=None,
        reference_database=None,
        default_set_aggregations=None,
        dec=6,
        max_rows=100,
        max_columns=20,):
  df = to_dataframe(iter_series, operator, start_year, end_year, reference_database, default_set_aggregations)
  df.style.set_properties(**{"text-align": "right", "precision": dec})
  with pd.option_context('display.max_rows', max_rows, 'display.max_columns', max_columns):  # more options can be specified also
    display(df)

def plot(iter_series,
         operator=None,
         start_year=None,
         end_year=None,
         reference_database=None,
         default_set_aggregations=None,
         layout={},
         **kwargs):
  df = to_dataframe(iter_series, operator, start_year, end_year, reference_database, default_set_aggregations)
  fig = df.plot(**kwargs)
  layout = {"yaxis": {"title": ""}, "xaxis": {"title": dt.TIME_AXIS_TITLE}, **layout}
  fig.update_layout(layout)
  return fig

def to_dataframe(iter_series, operator=None, start_year=None, end_year=None, reference_database=None, default_set_aggregations=None):
  if isinstance(iter_series, pd.Series):
    iter_series = [iter_series]

  iter_series = [aggregate_series(s, default_set_aggregations, reference_database) for s in iter_series]
  iter_series = compare_to_reference(iter_series, operator, reference_database)

  if start_year is None:
    start_year = dt.START_YEAR
  if end_year is None:
    end_year = dt.END_YEAR
  return merge_multiseries(*iter_series).loc[start_year:end_year]

def get_reference_database():
  if dt.REFERENCE_DATABASE is None:
    dt.REFERENCE_DATABASE = dt.Gdx(easygui.fileopenbox("Select reference gdx file", filetypes=["*.gdx"]))
  return dt.REFERENCE_DATABASE

def compare_to_reference(iter_series, operator, reference_database=None):
  """
  Applies an operator to each series in iter_series, comparing with the series with same name in the reference_database.
  """
  if not operator:
    return [s.copy() for s in iter_series]

  if reference_database is None:
    reference_database = get_reference_database()

  dimension_changed = [reference_database[s.name].index.nlevels != s.index.nlevels for s in iter_series]
  refs = [reference_database[s.name].loc[s.index] if not dimension_changed[i]
          else s.copy() * np.NaN
          for i, s in enumerate(iter_series)]

  if operator in ["q"]:
    return [s / b - 1 for s, b in zip(iter_series, refs)]
  elif operator in ["m"]:
    return [s - b for s, b in zip(iter_series, refs)]
  elif operator in ["s"]:
    for b in refs:
      b.name = "baseline." + b.name
    return [i for pair in zip(iter_series, refs) for i in pair]
  else:
    raise ValueError(f"{operator} is not a valid operator.")

def unstack_multiseries(series, keep_axis_index=-1):
  """
  Return a DataFrame from a Series.
  All levels of the Series are unstacked and concatenated as column names, except <keep_axis_index>.
  """
  if isinstance(series.index, pd.MultiIndex):
    series = series.copy()
    if series.name is None:
      series.name = ""
    series.index = pd.MultiIndex.from_tuples(flatten_keys(series.name, keys, keep_axis_index) for keys in series.index)
    df = series.unstack(0)[series.index.get_level_values(0).unique()]
  else:
    df = pd.DataFrame(series)
  return df

def flatten_keys(name, keys, keep_axis_index):
  keys_str = ','.join(map(str, keys[:keep_axis_index:]))
  flat_name = f"{name}[{keys_str}]"
  return flat_name, keys[keep_axis_index]

def merge_multiseries(*series, keep_axis_index=-1):
  """
  Return a DataFrame from any number of Series, with all levels except <keep_axis_index> concatenated as column names.
  """
  output = pd.DataFrame()
  for s in series:
    df = unstack_multiseries(s, keep_axis_index)
    for c in df.columns:
      output[c] = df[c]
  return output





import os

import numpy as np
import pandas as pd

from dreamtools import Gdx, GamsPandasDatabase


@np.vectorize
def approximately_equal(x, y, ndigits=0):
  return round(x, ndigits) == round(y, ndigits)


def test_gdx_read():
  gdx = Gdx("test.gdx")
  assert approximately_equal(gdx["qY"]["byg", 2010], 191)
  assert approximately_equal(gdx["qI_s"]["IB", "fre", 2010], 4.43)  # "IB" should be changed to "iB" in the GDX file.
  assert approximately_equal(gdx["eCx"], 1)
  assert gdx["s"].name == "s_"


def test_create_set_from_index():
  db = GamsPandasDatabase()
  t = pd.Index(range(2010, 2026), name="t")
  db.create_set("t", t)
  assert db["t"].name == "t"
  assert all(db["t"] == t)
  assert db.symbols["t"].domains_as_strings == ["*"]
  assert db.t.domains == ["*"]

  db.create_set("tsub", t[5:], domains=["t"])
  assert db["tsub"].name == "tsub"
  assert all(db["tsub"] == t[5:])
  assert db.symbols["tsub"].domains_as_strings == ["t"]
  assert db.tsub.domains == ["t"]

  s = pd.Index(["services", "goods"], name="s")
  st = pd.MultiIndex.from_product([s, t], names=["s", "t"])
  db.create_set("st", st)
  assert db["st"].name == "st"
  assert all(db["st"] == st[:])
  assert db.symbols["st"].domains_as_strings == ["s", "t"]
  assert db.st.domains == ["s", "t"]


def test_add_set_from_series():
  db = GamsPandasDatabase()
  t = pd.Index(range(2010, 2026), name="t")
  t = pd.Series([f"Year {i}" for i in t], index=t, name="t")
  db.add_set_from_series(t)
  assert db["t"].name == "t"
  assert all(db["t"].texts == t)
  assert all(db["t"] == t.index)
  assert db.symbols["t"].domains_as_strings == ["*"]

  tsub = t[5:]
  tsub.name = "tsub"
  db.add_set_from_series(tsub)
  assert db["tsub"].name == "tsub"
  assert all(db["tsub"].texts == tsub)
  assert all(db["tsub"] == tsub.index)
  assert db.symbols["tsub"].domains_as_strings == ["t"]
  assert db.tsub.domains == ["t"]

  s = pd.Index(["services", "goods"], name="s")
  st = pd.MultiIndex.from_product([s, t], names=["s", "t"])
  st = pd.Series([str(i) for i in st], index=st, name="st")
  db.add_set_from_series(st)
  assert db["st"].name == "st"
  assert all(db["st"].texts == st)
  assert all(db["st"] == st.index)
  assert db.symbols["st"].domains_as_strings == ["s", "t"]


def test_add_parameter_from_dataframe():
  db = GamsPandasDatabase()
  df = pd.DataFrame()
  df["t"] = range(2010, 2026)
  df["value"] = 1.3
  db.add_parameter_from_dataframe("par", df, add_missing_domains=True)
  assert all(db["par"] == 1.3)
  assert len(db["par"]) == 16


def test_multiply_added():
  db = GamsPandasDatabase()
  df = pd.DataFrame([
    [2010, "ser", 3],
    [2010, "goo", 2],
    [2020, "ser", 6],
    [2020, "goo", 4],
  ], columns=["t", "s", "value"])
  db.add_parameter_from_dataframe("q", df, add_missing_domains=True)

  df = pd.DataFrame([
    [2010, 1],
    [2020, 1.2],
  ], columns=["t", "value"])
  db.add_parameter_from_dataframe("p", df, add_missing_domains=True)

  v = db["p"] * db["q"]
  v.name = "v"
  db.add_parameter_from_series(v)

  assert db["v"][2020, "goo"] == 4.8


def test_add_parameter_from_series():
  db = GamsPandasDatabase()
  t = pd.Index(range(2010, 2026), name="t")
  par = pd.Series(1.4, index=t, name="par")
  db.add_parameter_from_series(par, add_missing_domains=True)
  assert all(db["par"] == 1.4)
  assert len(db["par"]) == 16

def test_create_variable():
  db = GamsPandasDatabase()
  db.create_variable("scalar", data=1)
  db.create_variable("vector", data=[1, 2])
  db.export("test_export.gdx")
  assert Gdx("test_export.gdx")["scalar"] == 1
  assert all(Gdx("test_export.gdx")["vector"] == [1, 2])

def test_add_variable_from_series():
  db = GamsPandasDatabase()
  t = pd.Index(range(2010, 2026), name="t")
  var = pd.Series(1.4, index=t, name="var")
  db.add_variable_from_series(var, add_missing_domains=True)
  assert all(db["var"] == 1.4)
  assert len(db["var"]) == 16

def test_add_variable_from_dataframe():
  db = GamsPandasDatabase()
  df = pd.DataFrame([
    [2010, "ser", 3],
    [2010, "goo", 2],
    [2020, "ser", 6],
    [2020, "goo", 4],
  ], columns=["t", "s", "value"])
  db.add_variable_from_dataframe("q", df, add_missing_domains=True)


def test_multiply_with_different_sets():
  assert approximately_equal(
    sum(Gdx("test.gdx")["qBNP"] * Gdx("test.gdx")["qI"] * Gdx("test.gdx")["qI_s"]),
    50730260150
  )


def test_export_with_no_changes():
  Gdx("test.gdx").export("test_export.gdx", relative_path=True)
  assert round(os.stat("test.gdx").st_size, -5) == round(os.stat("test_export.gdx").st_size, -5)


def test_export_variable_with_changes():
  gdx = Gdx("test.gdx")
  gdx["qY"] = gdx["qY"] * 2
  gdx.export("test_export.gdx", relative_path=True)
  old, new = Gdx("test.gdx"), Gdx("test_export.gdx")
  assert all(old["qY"] * 2 == new["qY"])


def test_export_scalar_with_changes():
  gdx = Gdx("test.gdx")
  gdx["eCx"] = gdx["eCx"] * 2
  gdx.export("test_export.gdx", relative_path=True)

  old, new = Gdx("test.gdx"), Gdx("test_export.gdx")
  assert approximately_equal(old["eCx"] * 2, new["eCx"])


def test_export_set_with_changes():
  gdx = Gdx("test.gdx")
  gdx["s"].texts["tje"] = "New text"
  gdx.export("test_export.gdx", relative_path=True)
  assert Gdx("test_export.gdx")["s"].texts["tje"] == "New text"


def test_copy_set():
  gdx = Gdx("test.gdx")
  gdx["alias"] = gdx["s"]
  gdx["alias"].name = "alias"
  index = gdx["alias"]
  domains = ["*" if i in (None, index.name) else i for i in gdx.get_domains_from_index(index, index.name)]
  gdx.database.add_set_dc(index.name, domains, "")
  index = index.copy()
  index.domains = domains
  gdx.series[index.name] = index
  gdx.export("test_export.gdx", relative_path=True)
  assert all(Gdx("test_export.gdx")["alias"] == gdx["s"])


def test_export_added_variable():
  gdx = Gdx("test.gdx")
  gdx.create_variable("foo", [gdx.a, gdx.t], explanatory_text="Variable added from Python.")
  gdx["foo"] = 42
  gdx.export("test_export.gdx", relative_path=True)
  assert all(approximately_equal(Gdx("test_export.gdx")["foo"], 42))

def test_export_NAs():
  db = GamsPandasDatabase()
  t = db.create_set("t", range(5))
  p = db.create_parameter("p", t)
  assert len(db["p"]) == 5
  assert len(db.symbols["p"]) == 0
  db.export("test_export.gdx")

  db = Gdx("test_export.gdx")
  assert all(pd.isna(db["p"]))
  assert len(db["p"]) == 0
  assert len(db.symbols["p"]) == 0

def test_detuple():
  assert GamsPandasDatabase.detuple("aaa") == "aaa"
  assert GamsPandasDatabase.detuple(("aaa",)) == "aaa"
  assert list(GamsPandasDatabase.detuple(("aaa", "bbb"))) == ["aaa", "bbb"]
  assert GamsPandasDatabase.detuple(1) == 1
  assert list(GamsPandasDatabase.detuple([1, 2])) == [1, 2]


def test_create_methods():
  # Create empty GamsPandasDatabase and alias creation methods
  db = GamsPandasDatabase()
  Par, Var, Set = db.create_parameter, db.create_variable, db.create_set

  # Create sets from scratch
  t = Set("t", range(2000, 2021), "Årstal")
  s = Set("s", ["tjenester", "fremstilling"], "Brancher", ["Tjenester", "Fremstilling"])
  st = Set("st", [s, t], "Branche x år dummy")
  sub = Set("sub", ["tjenester"], "Subset af brancher", domains=["s"])

  one2one = Set("one2one", [(2010, 2015), (2011, 2016)], "1 til 1 mapping", domains=["t", "t"])

  one2many = Set("one2many",
                 [("tot", "tjenester"), ("tot", "fremstilling")],
                 "1 til mange mapping", domains=["*", "s"],
                 )
  assert one2many.name == "one2many"
  assert one2many.names == ["*", "s"]
  assert one2many.domains == ["*", "s"]

  # Create parameters and variables based on zero ore more sets
  gq = Par("gq", None, "Produktivitets-vækst", 0.01)
  fq = Par("fp", t, "Vækstkorrektionsfaktor", (1 + 0.01)**(t-2010))
  d = Par("d", st, "Dummy")
  y = Var("y", [s,t], "Produktion")

  assert gq == 0.01
  assert all(fq.loc[2010:2011] == [1, 1.01])
  assert pd.isna(d["tjenester",2010])
  assert pd.isna(y["tjenester",2010])